import time
import pytest
import pytest_check as check

from typing import List, Dict
from multiprocessing import cpu_count
from aiohttp import ClientSession
import ssl
import certifi
import aresponses
import asyncio
import aiohttp
from aiohttp.web import BaseRequest

from rclib.multiprocessing.qpool import Orchestrator
from rclib.multiprocessing.tasks import Task, TaskState, RetryableTask
from rclib.multiprocessing.models import RequiresProps


# ======================================================================== #
# Helper Functions
# ======================================================================== #


def init_aiohttp_session():
    """Initialize an aiohttp session for workers."""
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    return ClientSession(connector=aiohttp.TCPConnector(ssl=ssl_context))


async def async_fetch_url(url: str, props: Dict[str, ClientSession]):
    """Perform asynchronous fetch using aiohttp."""
    try:
        async with props["aiohttp_session"].get(url) as resp:
            return resp.status
    except aiohttp.ClientError as e:
        print(f"Client error fetching {url}: {e}")
        return None
    except Exception as e:
        print(f"Error fetching {url}: {e}")
        return None


def fail():
    """Function that raises an exception."""
    raise Exception("This function always fails.")


# ======================================================================== #
# Shared Configuration
# ======================================================================== #

orchestrator_configs = {
    "async": {
        "actions": {"async_fetch_url": async_fetch_url},
        "props": {"aiohttp_session": init_aiohttp_session},
    },
    "fail": {
        "actions": {"fail": fail},
        "props": {},
    },
}


# ======================================================================== #
# Task Configuration
# ======================================================================== #


class Fail(Task):
    """Task that always fails."""

    action: str = "fail"


class AsyncFetchUrlTask(Task, RequiresProps):
    """Task for asynchronous URL fetching."""

    action: str = "async_fetch_url"
    requires_props: List[str] = ["aiohttp_session"]


# ======================================================================== #
# Mock Helper
# ======================================================================== #


def mocked_response(request: BaseRequest):
    time.sleep(1)
    return aresponses.Response(text="OK", status=200)


def setup_mock_responses(aresponses, urls: List[str]):
    """
    Centralize the setup of mock responses for aresponses.

    Args:
        aresponses: The aresponses instance.
        urls: List of URLs to mock.
    """
    for url in urls:
        hostname, path = "picsum.photos", f"/{url.split('/')[-1]}"
        aresponses.add(
            hostname,
            path,
            "GET",
            response=mocked_response,
        )


# ======================================================================== #
# Tests
# ======================================================================== #


@pytest.mark.asyncio
async def test_orchestrator_async(aresponses):
    # ======================================================================== #
    # A) Setup orchestrator and tasks
    # ======================================================================== #
    show_progress = False
    print(f"\nRunning test_orchestrator_async with show_progress={show_progress}...")

    sizes = [100, 200, 300, 400, 500]
    test_urls = [f"https://picsum.photos/{size}" for size in sizes]
    tasks = [AsyncFetchUrlTask(args=[url]) for url in test_urls]

    # Centralize mock response setup
    setup_mock_responses(aresponses, test_urls)

    orchestrator = Orchestrator(
        num_workers=cpu_count(),
        show_progress=show_progress,
        task_models=[AsyncFetchUrlTask],
        **orchestrator_configs["async"],
    )

    # ======================================================================== #
    # B) Run tasks and collect results
    # ======================================================================== #
    hold = orchestrator.add_tasks(tasks)
    start_time = time.monotonic()
    await hold
    job_results = await orchestrator.stop_workers()
    duration = time.monotonic() - start_time

    # ======================================================================== #
    # C) Validate results
    # ======================================================================== #
    check.equal(len(job_results), len(sizes), "Expected correct number of results")
    errors = sum(1 for result in job_results if result is None)
    check.equal(errors, 0, "Expected no errors in async approach")

    print(f"Test completed in {duration:.2f} seconds.")


@pytest.mark.asyncio
async def test_orchestrator_failure():
    # ======================================================================== #
    # A) Setup orchestrator and tasks
    # ======================================================================== #
    print("\nRunning test_orchestrator_failure...")

    tasks = [Fail() for _ in range(10)]

    orchestrator = Orchestrator(
        num_workers=cpu_count(),
        show_progress=False,
        task_models=[Fail],
        **orchestrator_configs["fail"],  # Inject shared actions and props
    )

    # ======================================================================== #
    # B) Run tasks and collect results
    # ======================================================================== #
    hold = orchestrator.add_tasks(tasks)
    await hold
    job_results = await orchestrator.stop_workers()

    # ======================================================================== #
    # C) Validate results
    # ======================================================================== #
    check.equal(len(job_results), 10, "Expected correct number of results")
    failures = [result.task.status == TaskState.fail for result in job_results]
    check.is_true(all(failures), "Expected all tasks to fail")


@pytest.mark.asyncio
async def test_add_tasks_of_wrong_model():
    # ======================================================================== #
    # A) Setup orchestrator and tasks
    # ======================================================================== #
    print("\nRunning test_add_tasks_of_wrong_model...")

    tasks = [AsyncFetchUrlTask(args=[None])]

    orchestrator = Orchestrator(
        num_workers=cpu_count(),
        show_progress=False,
        task_models=[Fail],
        **orchestrator_configs["fail"],  # Inject shared actions and props
    )

    # ======================================================================== #
    # B) Run tasks and collect results
    # ======================================================================== #
    enqueue_failures = await orchestrator.add_tasks(tasks)
    check.equal(len(enqueue_failures), 1, "Expected one enqueue failure")
    job_results = await orchestrator.stop_workers()

    # ======================================================================== #
    # C) Validate results
    # ======================================================================== #
    check.equal(len(job_results), 0, "Expected correct number of results")


@pytest.mark.asyncio
@pytest.mark.parametrize("show_progress", [True, False])
async def test_performance(aresponses, show_progress: bool):
    # ======================================================================== #
    # A) Setup orchestrator and tasks
    # ======================================================================== #
    print("\nRunning test_performance_comparison...")

    sizes = [[100] * 50, [200] * 50, [300] * 50, [400] * 50, [500] * 50]
    sizes = [item for sublist in sizes for item in sublist]
    test_urls = [f"https://picsum.photos/{size}" for size in sizes]
    tasks = [AsyncFetchUrlTask(args=[url]) for url in test_urls]
    results = {}

    # Centralize mock response setup
    setup_mock_responses(aresponses, test_urls)

    # ======================================================================== #
    # B) Baseline: synchronous fetching
    # ======================================================================== #
    start_time = time.monotonic()
    exceptions = 0
    for url in test_urls:
        try:
            async with init_aiohttp_session() as session:
                async with session.get(url) as resp:
                    text = await resp.text()
                    check.equal("OK", text, "Expected OK response")
                    check.equal(200, resp.status, "Expected 200 status")
        except Exception as e:
            print(f"Baseline error fetching {url}: {e}")
            exceptions += 1
    check.equal(exceptions, 0, "Expected no exceptions in baseline")
    baseline_duration = time.monotonic() - start_time

    results["no_pool"] = {
        "duration": baseline_duration,
        "items_per_second": len(test_urls) / baseline_duration,
    }

    # ======================================================================== #
    # C) Async Orchestrator
    # ======================================================================== #
    orchestrator = Orchestrator(
        num_workers=cpu_count(),
        show_progress=show_progress,
        task_models=[AsyncFetchUrlTask],
        **orchestrator_configs["async"],
    )

    hold = orchestrator.add_tasks(tasks)
    start_time_pool = time.monotonic()
    await hold
    job_results = await orchestrator.stop_workers()
    pool_duration = time.monotonic() - start_time_pool

    # ======================================================================== #
    # D) Validate results
    # ======================================================================== #
    check.equal(len(job_results), len(test_urls), "Expected correct number of results")
    exceptions = sum(1 for result in job_results if result is None)
    check.equal(exceptions, 0, "Expected no errors in pooled approach")

    results["pool"] = {
        "duration": pool_duration,
        "items_per_second": len(test_urls) / pool_duration,
    }

    # ======================================================================== #
    # E) Compare Results
    # ======================================================================== #
    for config, result in results.items():
        duration = result["duration"]
        ips = result["items_per_second"]
        speedup = baseline_duration / duration
        print(
            f"{config}: {duration:.2f} seconds, {ips:.2f} items/s, "
            f"{speedup:.2f}x speedup from baseline"
        )

    check.greater(
        results["pool"]["items_per_second"],
        results["no_pool"]["items_per_second"],
        "Expected the pooled approach to outpace the baseline",
    )


class RetryTask(RetryableTask):
    """Task that always fails and retries."""

    action: str = "fail"
    max_tries: int = 3


@pytest.mark.asyncio
async def test_task_retry(aresponses):
    # ======================================================================== #
    # A) Setup orchestrator and tasks
    # ======================================================================== #
    print("\nRunning test_task_retry...")

    tasks = [RetryTask() for _ in range(10)]

    orchestrator = Orchestrator(
        num_workers=cpu_count(),
        show_progress=False,
        task_models=[RetryTask],
        **orchestrator_configs["fail"],  # Inject shared actions and props
    )

    # ======================================================================== #
    # B) Run tasks and collect results
    # ======================================================================== #
    hold = orchestrator.add_tasks(tasks)
    await hold
    job_results = await orchestrator.stop_workers()

    # ======================================================================== #
    # C) Validate results
    # ======================================================================== #
    check.equal(len(job_results), 10, "Expected correct number of results")
    failures = [result.task.status == TaskState.fail for result in job_results]
    check.is_true(all(failures), "Expected all tasks to fail")
    retries = [result.task.tries == 3 for result in job_results]
    check.is_true(all(retries), "Expected all tasks to retry 3 times")
